function xn = bspline_transform_interp(x, ts, T, Llfactor)
%   Copyright (C) 2015 Javier Royuela del Val  <jroyval@lpi.tel.uva.es>
%                      Santiago Sanz Estébanez <ssanest@lpi.tel.uva.es>
%   Laboratorio de Procesado de Imagen, Universidad de Valladolid
%   www.lpi.tel.uva.es
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.


% tic;
%Llfactor = 2;
% T = sparse(T);
s = size(x);
L = s(3);
N = size(T,3);
E = ts.E;
Dp = ts.Dp;
% Transform region affected by control points
rp = (ts.E+1)*ts.Dp(:)/2;
%rp = [0 0];
n1 = Llfactor*(ts.v1(1)-rp(1)-1)+1 : Llfactor*(ts.v1(end)+rp(1)-1)+1;
n2 = Llfactor*(ts.v2(1)-rp(2)-1)+1 : Llfactor*(ts.v2(end)+rp(2)-1)+1;

% Avoid points outside image:
%    n1 = n1(0<n1 & n1<=size(x,2));
%    n2 = n2(0<n2 & n2<=size(x,1));

% Zero-matrix with control points displacement values positioned
Tpad = zeros([length(n2) length(n1) N L]);
Tpad(Llfactor*(ts.v2-ts.v2(1)+rp(2))+1,Llfactor*(ts.v1-ts.v1(1)+rp(1))+1,:,:) = T;

xh = -(E+1)/2 + 1/(Dp(1)*Llfactor) : 1/(Dp(1)*Llfactor) : (E+1)/2 - 1/(Dp(1)*Llfactor);
yh = -(E+1)/2 + 1/(Dp(2)*Llfactor) : 1/(Dp(2)*Llfactor) : (E+1)/2 - 1/(Dp(2)*Llfactor);
BB = bsplineN(yh.',E)*bsplineN(xh,E);

% xn => transformed coordinates = original + b-splines displacement
xn = repmat(permute(x,[1 2 4 3]),[1 1 N 1]);

parfor n=1:N
    x_ = x;
    for l=1:L
        xn(n2,n1,n,l) = x_(n2,n1,l) + conv2(Tpad(:,:,n,l),BB,'same');
    end
end
% toc;

% One line transform with convn
% (but conv2 is faster than convn)
% tic;
%   xn = repmat(permute(x,[1 2 4 3]),[1 1 N 1]);
%   xn(n1,n2,:,:) = bsxfun(@plus,xn(n1,n2,:,:),convnc(Tpad,BB,'same'));
% toc;

end