% Demo script to perform GWCS reconstruction of retrospectively
% undersamlped Cartesian data.
%
%	Copyright (C) 2015 Javier Royuela del Val <jroyval@lpi.tel.uva.es>
%	Laboratorio de Procesado de Imagen, Universidad de Valladolid
%	www.lpi.tel.uva.es
%
%   When indicated, some parts are based on code from Salman Asif:
%       http://users.ece.gatech.edu/sasif/dynamicMRI/index.html
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.

% Load data
load('gwcs_im.mat');
im = im./max(abs(im(:)));
[ROW, COL, N] = size(im);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% GROUPWISE REGISTRATION SETUP
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Original coordinates grid
clear x;
[x(:,:,1), x(:,:,2)] = meshgrid(1:COL,1:ROW);

MOT_WEIGHTS = [1, 6.5, 1, 15]; % Recularization parameters.
MOT_METRIC  = 'VI';            % Registration metric: Variance of the Intensity.
USE_GPU_REG = false;           % Use GPU for registration (only CUDA compatible).
ROI.type    = 'wholebox';      % Look for motion in the whole image.

 % LAUNCH REGISTRATION
fprintf('Registration\n');
regout = groupwise_reg(im, [], MOT_METRIC, MOT_WEIGHTS, USE_GPU_REG, ROI);
fprintf(1,' registration finished in %d seconds.\n', regout.exectime);
exectimes.registration = regout.exectime;

% Apply motion compensation to original image.
im_mc = interpolar(x,regout.xn,im,'cubic',0);
% Show video:
im_out = [im im_mc];
implay(im_out./max(abs(im_out(:))));
        