function [xn_back, xn_forw] = getBackAndForwardMotionFields(x, xn, xninv)
%getBackAndForwardMotionFields returns backward and forward motion fields
%   [xn_back, xn_forw] = getBackAndForwardMotionFields(x, xn, xninv) uses
%   the motion fields in xn and their inverse in xninv to generate the
%   backward and forward motion fields that describe the motion between
%   consecutive frames.
%
%   Copyright (C) 2015 Javier Royuela del Val  <jroyval@lpi.tel.uva.es>
%                      Santiago Sanz Estébanez <ssanest@lpi.tel.uva.es>
%   Laboratorio de Procesado de Imagen, Universidad de Valladolid
%   www.lpi.tel.uva.es
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.



sx = size(xn);
x = gather(x);
xn = gather(xn);
xn_back = zeros(sx);
xn_forw = zeros(sx);
N = sx(3);
for i=1:N
    ileft  = mod(i-2, N) + 1;
    iright = mod(i  , N) + 1;

    xn_back(:,:,i,1)  = interp2(x(:,:,1), x(:,:,2), xn(:,:,iright,1), xninv(:,:,i,1),  xninv(:,:,i,2),'*cubic',0);
    xn_back(:,:,i,2)  = interp2(x(:,:,1), x(:,:,2), xn(:,:,iright,2), xninv(:,:,i,1),  xninv(:,:,i,2),'*cubic',0);
    xn_forw(:,:,i,1)  = interp2(x(:,:,1), x(:,:,2), xn(:,:,ileft ,1), xninv(:,:,i,1),  xninv(:,:,i,2),'*cubic',0);
    xn_forw(:,:,i,2)  = interp2(x(:,:,1), x(:,:,2), xn(:,:,ileft ,2), xninv(:,:,i,1),  xninv(:,:,i,2),'*cubic',0);
end



end

