function xn_inv = inv_motion_field( xn , ts, T_ite, varargin)
%UNTITLED Summary of this function goes here
%   Detailed explanation goes here
%   Inv motion field numericacly by min ||T^-1(T(x)) - x||^2
%   Metz et al 2011
%
%   Copyright (C) 2015 Javier Royuela del Val  <jroyval@lpi.tel.uva.es>
%                      Santiago Sanz Estébanez <ssanest@lpi.tel.uva.es>
%   Laboratorio de Procesado de Imagen, Universidad de Valladolid
%   www.lpi.tel.uva.es
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.


    lambda = 0.1;
    thres = 1e-5;
    max_iter = 500;
    Lfun = 1;
    if nargin>1
        for ii = 1:length(varargin)
            if strcmpi(varargin{ii},'lambda')
                lambda = varargin{ii+1};
            elseif strcmpi(varargin{ii},'threshold')
                thres = varargin{ii+1};
            elseif strcmpi(varargin{ii},'maxiter')
                max_iter = varargin{ii+1};
            elseif strcmpi(varargin{ii},'fun')
                Lfun = varargin{ii+1};
            end
        end
    end

    s = size(xn);
    L = s(end);

    switch L
        case 2
            if Lfun == 1
%                 xn_inv = backwards2forwards_n(xn);
%                 xn_inv = inv_2D(xn,[],lambda,thres,max_iter,ts,T_ite);
                xn_inv = inv_2D_old(xn,lambda,thres,max_iter);
            elseif Lfun == 2
                xn_inv = inv_2D_old_2(xn,ts,T_ite,lambda,thres,max_iter);
            else
                xn_inv = inv_2D_lsqnonlin(xn,ts,T_ite,lambda,thres,max_iter);
            end

        case 3
            xn_inv = inv_3D(xn,lambda,thres,max_iter);
    end

end

function xn = backwards2forwards_n(xn)
    parfor i=1:size(xn,3)
        xn(:,:,i,:)=backwards2forwards(squeeze(xn(:,:,i,:)));
    end
end

function xn_inv = inv_2D(xn,xn_inv,lambda,thres,max_iter,ts,T_ite)
    s = size(xn);

    ROW = s(1);
    COL = s(2);
    N   = s(3);
    
    [x(:,:,1), x(:,:,2)] = meshgrid(1:COL,1:ROW);
    xn = gather(xn);
    
    % Initial solution:
    if ~exist('xn_inv','var') || isempty(xn_inv)
        xn_inv(:,:,:,1) = repmat(permute(x(:,:,1),[1 2 4 3]),[1 1 N,1]);
        xn_inv(:,:,:,2) = repmat(permute(x(:,:,2),[1 2 4 3]),[1 1 N,1]);
    end
    
    opts.t0 = lambda;
    opts.p2 = thres;
    opts.p4 = thres;
    opts.max_iter = max_iter;
    opts.ptr = true;
    opts.verbose = true;

    
    xn_inv = cgsolver(xn_inv, @fun, opts, x, xn, ts, T_ite);
    
end

function [f, g] = fun(xn_inv, x, xn)
    % xn  = varargin{1};
    s = size(xn);
    g = zeros(s);
    s_ = [s(1) s(2) 1 2];
    
    parfor i=1:s(3)
        g_=zeros(s_);
        x_ = x;
        xn_ = xn(:,:,i,:);
        xn_inv_ = xn_inv(:,:,i,:);
        g_(:,:,1,1) = interp2(x_(:,:,1),x_(:,:,2),xn_(:,:,1,1),xn_inv_(:,:,1,1),xn_inv_(:,:,1,2),'*cubic')-x_(:,:,1);
        g_(:,:,1,2) = interp2(x_(:,:,1),x_(:,:,2),xn_(:,:,1,2),xn_inv_(:,:,1,1),xn_inv_(:,:,1,2),'*cubic')-x_(:,:,2);
        g(:,:,i,:)=g_;
    end
    
    f = 0.5*sum(g(:).^2);

end

function xn_inv = inv_2D_old(xn,lambda,thres,max_iter)
   
    s = size(xn);

    ROW = s(1);
    COL = s(2);
    N   = s(3);
    
    [x(:,:,1), x(:,:,2)] = meshgrid(1:COL,1:ROW);
    xn = gather(xn);
    n = ROW*COL*N;

    stop_criteria = false;

    xn_inv(:,:,:,1) = repmat(permute(x(:,:,1),[1 2 4 3]),[1 1 N,1]);
    xn_inv(:,:,:,2) = repmat(permute(x(:,:,2),[1 2 4 3]),[1 1 N,1]);

    [f, g] = fun(xn_inv,x,xn);
    
    Dif = -g;
    iter=0;
    while ~stop_criteria && iter < max_iter
        iter = iter + 1;

        % Derivada direccional: proyeccion gradiente sobre Dif
        g0 = Dif(:).'*g(:); 
        if g0>=0 %No direccion de descenso
            fprintf(1,'%d: warn: No direccion de descenso\n',iter);
            %break;
        end

        gold = g;
        [f, g] = fun(xn_inv,x,xn);

        % Actualizar punto:
        xn_inv = xn_inv + lambda*Dif;

        gnorm = sum(g(:).^2);
        
        if sqrt(abs(gnorm)) < thres * n
            stop_criteria = true;
        end

        % Gradiente conjugado
        y = g - gold;
        bgold = gold(:)'*gold(:) + eps;
        % Polak-Ribiere-Polyak variant
        bpr = g(:)'*y(:)/bgold;

        % Fletcher-Reeves variant
        bfr = g(:)'*g(:)/bgold;
        % Combined variant
        if bpr < -bfr
            bk = -bfr;
        elseif bpr> bfr
            bk = bfr;
        else
            bk = bpr;
        end
        
        Dif = bk*Dif - g;
        
        if (mod(iter,1)==0)
            fprintf('Iter %d, f = %f, g = %f\n',iter,f,sqrt(gnorm));
        end
    end
end

function [f ,g] = funfg(x,y,xn,Jxn)
    s=size(xn);
    n = prod(s(1:3));
    
    Ty = zeros(size(xn));
    Jy = zeros(size(Jxn));
    for i=1:s(3)
        Ty(:,:,i,1)   = interp2(x(:,:,1),x(:,:,2),xn(:,:,i,1),y(:,:,i,1),y(:,:,i,2),'*cubic')-x(:,:,1);
        Ty(:,:,i,2)   = interp2(x(:,:,1),x(:,:,2),xn(:,:,i,2),y(:,:,i,1),y(:,:,i,2),'*cubic')-x(:,:,2);
        
        Jy(:,:,i,1,1) = interp2(x(:,:,1),x(:,:,2),Jxn(:,:,i,1,1),y(:,:,i,1),y(:,:,i,2),'*cubic');
        Jy(:,:,i,1,2) = interp2(x(:,:,1),x(:,:,2),Jxn(:,:,i,1,2),y(:,:,i,1),y(:,:,i,2),'*cubic');
        Jy(:,:,i,2,1) = interp2(x(:,:,1),x(:,:,2),Jxn(:,:,i,2,1),y(:,:,i,1),y(:,:,i,2),'*cubic');
        Jy(:,:,i,2,2) = interp2(x(:,:,1),x(:,:,2),Jxn(:,:,i,2,2),y(:,:,i,1),y(:,:,i,2),'*cubic');
    end
    
    g = zeros(size(xn));
    g(:,:,:,1) = Ty(:,:,:,1).*Jy(:,:,:,1,1) + Ty(:,:,:,2).*Jy(:,:,:,2,1);
    g(:,:,:,2) = Ty(:,:,:,1).*Jy(:,:,:,1,2) + Ty(:,:,:,2).*Jy(:,:,:,2,2);
    
    f = 0.5*sum(Ty(:).^2);
       
end

function f = funfvec(x,y,xn,xi)
    s=size(xn);
%     s = [s 1 1];
    y = reshape(y,[1 1 s(3) 2]);
    f = zeros(size(y));
    for i=1:s(3)
        for l=1:s(4)
            f(:,:,i,l) = interp2(x(:,:,1),x(:,:,2),xn(:,:,i,l),y(:,:,i,1),y(:,:,i,2),'*cubic')-xi(:,:,l);
        end
    end
    f = f(:);
end

function xn_inv = inv_2D_lsqnonlin(xn,ts,T_ite,lambda,thres,max_iter)
    s = size(xn);
    [x(:,:,1), x(:,:,2)] = meshgrid(1:s(2),1:s(1));
    
    x0 = permute(x,[1 2 4 3]);
     
    options = optimoptions(@lsqnonlin,'Algorithm','levenberg-marquardt');
    for i=1:s(3)
        for i1 = 1:s(1)
            for i2 = 1:s(2)
                fun = @(y)funfvec(x,y,xn(:,:,i,:),x(i1,i2,:));
                [xn_inv(i1,i2,i,:), resnorm] = lsqnonlin(fun, vec(x0(i1,i2,:)), [], [], options);
            end
        end
    end
end

function xn_inv = inv_2D_old_2(xn,ts,T_ite,lambda,thres,max_iter)
   
    s = size(xn);

    ROW = s(1);
    COL = s(2);
    N   = s(3);
    [x(:,:,1), x(:,:,2)] = meshgrid(1:COL,1:ROW);
    xn          = bspline_transform(x,ts,T_ite);
    [Jx(:,:,:,:,1), Jx(:,:,:,:,2)]  = bspline_spatial_gradient(x,ts,T_ite,1);
    
    %Jx(:,:,:,i,j) = d(Tj-xj)/dxi 
    fun = @(xninv)funfg(x,xninv,xn,Jx);
    
    n = ROW*COL*N;

    stop_criteria = false;

    xn_inv(:,:,:,1) = repmat(permute(x(:,:,1),[1 2 4 3]),[1 1 N,1]);
    xn_inv(:,:,:,2) = repmat(permute(x(:,:,2),[1 2 4 3]),[1 1 N,1]);

    [~, g] = fun(xn_inv);
    Dif = -g;
    iter=0;
    while ~stop_criteria && iter < max_iter
        iter = iter + 1;

        % Derivada direccional: proyeccion gradiente sobre Dif
        g0 = Dif(:).'*g(:); 
        if g0>=0 %No direccion de descenso
            fprintf(1,'%d: warn: No direccion de descenso\n',iter);
            %break;
        end

%         gold = g;
        [f, g] = fun(xn_inv);

        % Actualizar punto:
        xn_inv = xn_inv + lambda*Dif;

        gnorm = sum(g(:).^2);
        
        if sqrt(abs(gnorm)) < thres * n
            stop_criteria = true;
        end

        % Gradiente conjugado
%         y = g - gold;
%         bgold = gold(:)'*gold(:) + eps;
%         % Polak-Ribiere-Polyak variant
%         bpr = g(:)'*y(:)/bgold;
% 
%         % Fletcher-Reeves variant
%         bfr = g(:)'*g(:)/bgold;
%         % Combined variant
%         if bpr < -bfr
%             bk = -bfr;
%         elseif bpr> bfr
%             bk = bfr;
%         else
%             bk = bpr;
%         end
%         
%         Dif = bk*Dif - g;
        Dif = -g;
        
        if (mod(iter,1)==0)
            fprintf('Iter %d, f = %f, g = %f\n',iter,f,sqrt(gnorm));
        end
    end
end

function x_inv = inv_3D(xn,lambda,thres,max_iter)
   
    s = size(xn);

    ROW = s(1);
    COL = s(2);
    SLC = s(3);
    N = s(4);
    
    [x1, x2, x3] = meshgrid(1:COL,1:ROW,1:SLC);
    
    xn  = gather(xn);
    xn1 = xn(:,:,:,:,1);
    xn2 = xn(:,:,:,:,2);
    xn3 = xn(:,:,:,:,3);
    clear xn;

    n = ROW*COL*SLC*N;

    stop_criteria = false;

    xn_inv1 = repmat(permute(x1,[1 2 3 5 4]),[1 1 1 N 1]);
    xn_inv2 = repmat(permute(x2,[1 2 3 5 4]),[1 1 1 N 1]);
    xn_inv3 = repmat(permute(x3,[1 2 3 5 4]),[1 1 1 N 1]);
    g1 = zeros([ROW COL SLC N]);
    g2 = zeros([ROW COL SLC N]);
    g3 = zeros([ROW COL SLC N]);
    
    parfor i=1:N
        g1(:,:,:,i) = interp3_gpu(x1,x2,x3,xn1(:,:,:,i),xn_inv1(:,:,:,i),xn_inv2(:,:,:,i),xn_inv3(:,:,:,i),'*linear',0)-x1;
        g2(:,:,:,i) = interp3_gpu(x1,x2,x3,xn2(:,:,:,i),xn_inv1(:,:,:,i),xn_inv2(:,:,:,i),xn_inv3(:,:,:,i),'*linear',0)-x2;
        g3(:,:,:,i) = interp3_gpu(x1,x2,x3,xn3(:,:,:,i),xn_inv1(:,:,:,i),xn_inv2(:,:,:,i),xn_inv3(:,:,:,i),'*linear',0)-x3;
    end

    Dif1 = -g1;
    Dif2 = -g2;
    Dif3 = -g3;

    iter=0;
    while ~stop_criteria && iter < max_iter
        iter = iter + 1;

        % Derivada direccional: proyeccion gradiente sobre Dif
        g0 = Dif1(:).'*g1(:) + Dif2(:).'*g2(:) + Dif3(:).'*g3(:); 
        if g0>=0 %No direccion de descenso
            fprintf(1,'%d: warn: No direccion de descenso\n',iter);
            %break;
        end

        gold1 = g1;
        gold2 = g2;
        gold3 = g3;
        parfor i=1:N
            g1(:,:,:,i) = interp3_gpu(x1,x2,x3,xn1(:,:,:,i),xn_inv1(:,:,:,i),xn_inv2(:,:,:,i),xn_inv3(:,:,:,i),'*linear',0)-x1;
            g2(:,:,:,i) = interp3_gpu(x1,x2,x3,xn2(:,:,:,i),xn_inv1(:,:,:,i),xn_inv2(:,:,:,i),xn_inv3(:,:,:,i),'*linear',0)-x2;
            g3(:,:,:,i) = interp3_gpu(x1,x2,x3,xn3(:,:,:,i),xn_inv1(:,:,:,i),xn_inv2(:,:,:,i),xn_inv3(:,:,:,i),'*linear',0)-x3;
        end

        % Actualizar punto:
        % xn_inv_old = xn_inv;
        xn_inv1 = xn_inv1 +lambda*Dif1;
        xn_inv2 = xn_inv2 +lambda*Dif2;
        xn_inv3 = xn_inv3 +lambda*Dif3;

        gnorm1 = sum(g1(:).^2);
        gnorm2 = sum(g2(:).^2);
        gnorm3 = sum(g3(:).^2);
        if sqrt(abs(gnorm1+gnorm2+gnorm3)) < thres * n
            stop_criteria = true;
        else
            iter = iter + 1;
        end

        % Gradiente conjugado
        y1     = g1 - gold1;
        y2     = g2 - gold2;
        y3     = g3 - gold3;
        bgold1 = (gold1(:)'*gold1(:)+eps);
        bgold2 = (gold2(:)'*gold2(:)+eps);
        bgold3 = (gold3(:)'*gold3(:)+eps);
        % Polak-Ribiere-Polyak variant
        bpr1 = g1(:)'*y1(:)/bgold1;
        bpr2 = g2(:)'*y2(:)/bgold2;
        bpr3 = g3(:)'*y3(:)/bgold3;

        % Fletcher-Reeves variant
        bfr1 = g1(:)'*g1(:)/bgold1;
        bfr2 = g2(:)'*g2(:)/bgold2;
        bfr3 = g3(:)'*g3(:)/bgold3;

        % Combined variant
        if bpr1 < -bfr1
            bk1 = -bfr1;
        %    fprintf(1,'-');
        elseif bpr1> bfr1
            bk1 = bfr1;
        %    fprintf(1,'+');
        else
            bk1 = bpr1;
        end
        % Combined variant
        if bpr2 < -bfr2
            bk2 = -bfr2;
        %    fprintf(1,'-');
        elseif bpr2> bfr2
            bk2 = bfr2;
        %    fprintf(1,'+');
        else
            bk2 = bpr2;
        end
        % Combined variant
        if bpr3 < -bfr3
            bk3 = -bfr3;
        %    fprintf(1,'-');
        elseif bpr3> bfr3
            bk3 = bfr3;
        %    fprintf(1,'+');
        else
            bk3 = bpr3;
        end

        Dif1 = bk1 * Dif1 + g1;
        Dif2 = bk2 * Dif2 + g2;
        Dif3 = bk3 * Dif3 + g3;

        if (mod(iter,1)==0)
            fprintf('Iter %d, f = %f\n',iter,sqrt(gnorm1 + gnorm2 + gnorm3));
        end
    end

    x_inv = cat(5,xn_inv1,xn_inv2,xn_inv3);
end